# -*- coding: utf-8 -*-

import numpy as np


def get_stratey_for_corrupting(corrupt_suject):
    if corrupt_suject:
        start_of_columns_to_maintain = 1
        end_of_columns_to_maintain = 3
        start_of_corrupted_colum = 0
        end_of_corrupted_column = 1

        concatenate_fct = _concatenate_entites_first

    else:
        start_of_columns_to_maintain = 0
        end_of_columns_to_maintain = 2
        start_of_corrupted_colum = 2
        end_of_corrupted_column = 3

        concatenate_fct = _concatenate_entites_last

    return (start_of_columns_to_maintain, end_of_columns_to_maintain), (
        start_of_corrupted_colum, end_of_corrupted_column), concatenate_fct


def _concatenate_entites_first(candidate_entities, tuples):
    candidate_entities = np.reshape(candidate_entities, newshape=(-1, 1))
    return np.concatenate([candidate_entities, tuples], axis=1)


def _concatenate_entites_last(candidate_entities, tuples):
    candidate_entities = np.reshape(candidate_entities, newshape=(-1, 1))
    return np.concatenate([tuples, candidate_entities], axis=1)



# -*- coding: utf-8 -*-

"""Utilities for getting and initializing KGE models."""

from typing import Dict

from torch.nn import Module

from pykeen.constants import (
    CONV_E_NAME, DISTMULT_NAME, ERMLP_NAME, KG_EMBEDDING_MODEL_NAME, RESCAL_NAME, SE_NAME, TRANS_D_NAME, TRANS_E_NAME,
    TRANS_H_NAME, TRANS_R_NAME, UM_NAME,
)
from pykeen.kge_models import (
    ConvE, DistMult, ERMLP, RESCAL, StructuredEmbedding, TransD, TransE, TransH, TransR, UnstructuredModel,
)

__all__ = [
    'KGE_MODELS',
    'get_kge_model',
]

KGE_MODELS = {
    TRANS_E_NAME: TransE,
    TRANS_H_NAME: TransH,
    TRANS_D_NAME: TransD,
    TRANS_R_NAME: TransR,
    SE_NAME: StructuredEmbedding,
    UM_NAME: UnstructuredModel,
    DISTMULT_NAME: DistMult,
    ERMLP_NAME: ERMLP,
    RESCAL_NAME: RESCAL,
    CONV_E_NAME: ConvE,
}


def get_kge_model(config: Dict) -> Module:
    """Get an instance of a knowledge graph embedding model with the given configuration."""
    kge_model_name = config[KG_EMBEDDING_MODEL_NAME]
    kge_model_cls = KGE_MODELS.get(kge_model_name)

    if kge_model_cls is None:
        raise ValueError(f'Invalid KGE model name: {kge_model_name}')

    return kge_model_cls(config=config)



# -*- coding: utf-8 -*-

"""Helper script to query parameters."""

import json
import os
from collections import OrderedDict

import click
from prompt_toolkit import prompt

from pykeen.constants import (
    BINARY_QUESTION_MAPPING, CONFIG_FILE_ERROR_MSG, CONFIG_FILE_PROMPT_MSG, CPU, GPU, HPO_MODE, ID_TO_KG_MODEL_MAPPING,
    ID_TO_OPTIMIZER_MAPPING, KG_MODEL_TO_ID_MAPPING, OPTIMIZER_TO_ID_MAPPING, PYKEEN, TRAINING_MODE,
)


def _is_correct_format(path):
    if path.startswith('ndex:') or path.endswith('.tsv') or path.endswith('.nt'):
        return True
    else:
        return False

def get_input_path(prompt_msg, error_msg, is_dataset=False):
    while True:
        user_input = prompt(prompt_msg, ).strip('"').strip("'")

        if not os.path.exists(os.path.dirname(user_input)):
            click.echo(error_msg)
            continue
        if is_dataset:
            if not _is_correct_format(path=user_input):
                click.echo()
                click.echo('Invalid data source, following data sources are supported:\n'
                           'A string path to a .tsv file containing 3 columns corresponding to subject, predicate, and object.\n'
                           'A string path to a .nt RDF file serialized in N-Triples format.\n'
                           'A string NDEx network UUID prefixed by "ndex:" like in ndex:f93f402c-86d4-11e7-a10d-0ac135e8bacf')
                click.echo()
                continue

        return user_input


def select_keen_execution_mode(lib_name=PYKEEN):
    r = click.confirm('Do you have hyper-parameters? If not, %s will be configured for hyper-parameter search.' % (lib_name),
                      default=False)
    return TRAINING_MODE if r else HPO_MODE


def select_embedding_model():
    click.echo('Please select the embedding model you want to train:')
    for model, id in KG_MODEL_TO_ID_MAPPING.items():
        click.echo("%s: %s" % (model, id))

    ids = list(KG_MODEL_TO_ID_MAPPING.values())
    available_models = list(KG_MODEL_TO_ID_MAPPING.keys())

    while True:
        user_input = prompt('> Please select one of the options: ')

        if user_input not in ids:
            click.echo(
                "Invalid input, please type in a number between %s and %s indicating the model id.\n"
                "For example type %s to select the model %s and press enter" % (
                    available_models[0], ids[0], ids[0], available_models[0]))
            click.echo()
        else:
            return ID_TO_KG_MODEL_MAPPING[user_input]


def select_integer_value(print_msg, prompt_msg, error_msg):
    click.echo(print_msg)

    while True:
        user_input = prompt(prompt_msg)

        if user_input.isnumeric():
            return int(user_input)

        click.echo(error_msg)


def select_float_value(print_msg, prompt_msg, error_msg):
    click.echo(print_msg)

    while True:
        user_input = prompt(prompt_msg)
        try:
            float_value = float(user_input)
            return float_value
        except ValueError:
            click.echo(error_msg)


def select_zero_one_float_value(print_msg, prompt_msg, error_msg):
    click.echo(print_msg)

    while True:
        user_input = prompt(prompt_msg)
        try:
            float_value = float(user_input)
            if not (0 <= float_value <= 1):
                continue
            return float_value
        except ValueError:
            click.echo(error_msg)


def ask_for_evaluation():
    click.echo('Do you want to evaluate your model?')

    while True:
        user_input = prompt('> Please type \'yes\' or \'no\': ')
        if user_input != 'yes' and user_input != 'no':
            click.echo('Invalid input, please type \'yes\' or \'no\' and press enter.\n'
                       'If you type \'yes\' it means that you want to evaluate your model after it is trained.')
        else:
            return BINARY_QUESTION_MAPPING[user_input]


def ask_for_test_set():
    click.echo('Do you provide a test set yourself?')

    while True:
        user_input = prompt('> Please type \'yes\' or \'no\': ')

        if user_input != 'yes' and user_input != 'no':
            click.echo('Invalid input, please type \'yes\' or \'no\' and press enter.\n'
                       'If you type \'yes\' it means that you provide a test set yourself.')
        else:
            return BINARY_QUESTION_MAPPING[user_input]


def select_ratio_for_test_set():
    while True:
        user_input = prompt('> Please select the ratio: ')

        try:
            ratio = float(user_input)
            if 0. < ratio < 1.:
                return ratio
        except ValueError:
            pass

        click.echo('Invalid input, the ratio should be 0.< ratio < 1. (e.g. 0.2).\n'
                   'Please try again.')


def select_preferred_device():
    click.secho(click.style("Current Step: Please specify tthe preferred device (GPU or CPU).", fg='blue'))

    while True:
        user_input = prompt('> Please type \'GPU\' or \'CPU\': ').lower()
        if user_input == GPU or user_input == CPU:
            return user_input
        else:
            click.echo('Invalid input, please type in \'GPU\' or \'CPU\' and press enter.')


def ask_for_filtering_of_negatives():
    click.echo('Do you want to filter out negative triples during evaluation of your model?')

    while True:
        user_input = prompt('> Please type \'yes\' or \'no\': ')

        if user_input != 'yes' and user_input != 'no':
            click.echo('Invalid input, please type \'yes\' or \'no\' and press enter.\n'
                       'If you type \'yes\' it means that you provide a test set yourself.')
        else:
            return BINARY_QUESTION_MAPPING[user_input]


def load_config_file():
    path_to_config_file = get_input_path(prompt_msg=CONFIG_FILE_PROMPT_MSG, error_msg=CONFIG_FILE_ERROR_MSG)
    while True:
        with open(path_to_config_file, 'rb') as f:
            try:
                config = json.load(f)
                assert type(config) == dict or type(config) == OrderedDict
                return config
            except:
                click.echo('Invalid file, the configuration must be a JSON file.\n'
                           'Please try again.')
                path_to_config_file = get_input_path(prompt_msg=CONFIG_FILE_PROMPT_MSG, error_msg=CONFIG_FILE_ERROR_MSG)


def ask_for_existing_config_file():
    click.echo('Do you provide an existing configuration file?\n')

    while True:
        user_input = prompt('> Please type \'yes\' or \'no\': ')

        if user_input != 'yes' and user_input != 'no':
            click.echo('Invalid input, please type \'yes\' or \'no\' and press enter.\n'
                       'If you type \'yes\' it means that you provide a configuration file.')
        else:
            return BINARY_QUESTION_MAPPING[user_input]


def query_output_directory():
    click.echo('Please provide the path to your output directory.\n')
    click.echo()

    while True:
        user_input = prompt('> Path to output directory:')
        if os.path.exists(os.path.dirname(user_input)):
            return user_input
        else:
            click.echo('Invalid input, please make sure that the path to the directory exists.\n'
                       'Please try again.')


def query_height_and_width_for_conv_e(embedding_dim):
    click.echo("Note: Height and width must be positive positive integers.\n"
               "Besides, height * width must equal to  embedding dimension \'%d\'" % embedding_dim)
    click.echo()

    while True:
        height = prompt('> Height:')

        if not height.isnumeric():
            click.echo("Invalid input, please make sure that height is a positive integer.")
            continue

        width = prompt('> Width:')

        if not width.isnumeric():
            click.echo("Invalid input, please make sure that height is a positive integer.")
            continue

        if not (int(height) * int(width) == embedding_dim):
            click.echo("Invalid input, height * width are not equal to \'%d\' (your specified embedding dimension).\n"
                       "Please try again, and fulfill the constraint)" % embedding_dim)
        else:
            return int(height), int(width)


def query_kernel_param(depending_param, print_msg, prompt_msg, error_msg):
    click.echo(print_msg % depending_param)

    while True:
        kernel_param = prompt(prompt_msg)

        if not (kernel_param.isnumeric() and int(kernel_param) <= depending_param):
            click.echo(error_msg % depending_param)
        else:
            return int(kernel_param)


def select_float_values(print_msg, prompt_msg, error_msg):
    click.echo(print_msg)
    float_values = []
    is_valid_input = False

    while not is_valid_input:
        user_input = prompt(prompt_msg)
        user_input = user_input.split(',')
        is_valid_input = True

        for float_value in user_input:
            try:
                float_value = float(float_value)
                float_values.append(float_value)
            except ValueError:
                click.echo(error_msg)
                is_valid_input = False
                break

    return float_values


def select_zero_one_range_float_values(print_msg, prompt_msg, error_msg):
    click.echo(print_msg)
    float_values = []
    is_valid_input = False

    while not is_valid_input:
        user_input = prompt(prompt_msg)
        user_input = user_input.split(',')
        is_valid_input = True

        for float_value in user_input:
            try:
                float_value = float(float_value)
            except ValueError:
                click.echo(error_msg)
                is_valid_input = False
                break

            if 0. <= float_value <= 1.:
                print("hey")
                float_values.append(float_value)
            else:
                click.echo(error_msg)
                is_valid_input = False
                break

    return float_values


def select_positive_integer_values(print_msg, prompt_msg, error_msg):
    click.echo(print_msg)
    integers = []
    is_valid_input = False

    while not is_valid_input:
        user_input = prompt(prompt_msg)
        user_input = user_input.split(',')
        is_valid_input = True

        for integer in user_input:
            if integer.isnumeric():
                integers.append(int(integer))
            else:
                click.echo(error_msg)
                is_valid_input = False
                break

    return integers


def select_optimizer():
    click.echo('Please select the optimizer you want to train your model with:')
    for optimizer, id in OPTIMIZER_TO_ID_MAPPING.items():
        click.echo("%s: %s" % (optimizer, id))

    ids = list(OPTIMIZER_TO_ID_MAPPING.values())
    available_optimizers = list(OPTIMIZER_TO_ID_MAPPING.keys())

    while True:
        user_input = prompt('> Please select one of the options: ')

        if user_input not in ids:
            click.echo(
                "Invalid input, please type in a number between %s and %s indicating the optimizer id.\n"
                "For example type %s to select the model %s and press enter" % (
                    available_optimizers[0], ids[0], ids[0], available_optimizers[0]))
            click.echo()
        else:
            return ID_TO_OPTIMIZER_MAPPING[user_input]


def select_heights_and_widths(embedding_dimensions):
    heights = []
    widths = []

    for embedding_dim in embedding_dimensions:
        is_valid_input = False
        while not is_valid_input:
            print("Specify height for specified embedding dimension %d ." % embedding_dim)
            height = prompt('> Height:')

            print("Specify width for specified embedding dimension %d ." % embedding_dim)
            width = prompt('> Width:')

            if not (height.isnumeric() and width.isnumeric() and int(height) * int(width) == embedding_dim):
                print("Invalid input, height and width must be positive integers, and height * width must equal the "
                      " specified embedding dimension of \'%d\'." % embedding_dim)
            else:
                heights.append(int(height))
                widths.append(int(width))
                is_valid_input = True
        print()

    return heights, widths


def select_kernel_sizes(depending_params, print_msg, prompt_msg, error_msg):
    kernel_params = []
    print(print_msg)

    for dep_param in depending_params:
        is_valid_input = False

        while not is_valid_input:
            kernel_param = prompt(prompt_msg % dep_param)

            if not (kernel_param.isnumeric() and int(kernel_param) <= dep_param):
                print(error_msg % dep_param)
            else:
                kernel_params.append(int(kernel_param))
                is_valid_input = True
        print()

    return kernel_params



# -*- coding: utf-8 -*-

"""Script for predicting links based on a trained model."""

import json
import os

import numpy as np
import torch
from torch.nn import Module

from pykeen.constants import CPU, GPU, PREFERRED_DEVICE
from pykeen.kge_models import get_kge_model
from pykeen.utilities.prediction_utils import make_predictions


def start_predictions_pipeline(model_direc: str, data_direc: str):
    # Load configuration file
    with open(os.path.join(model_direc, 'configuration.json')) as f:
        config = json.load(f)

    # Load entity to id mapping
    with open(os.path.join(model_direc, 'entity_to_id.json')) as f:
        entity_to_id = json.load(f)

    # Load relation to id mapping
    with open(os.path.join(model_direc, 'relation_to_id.json')) as f:
        relation_to_id = json.load(f)

    trained_kge_model: Module = get_kge_model(config=config)
    path_to_model = os.path.join(model_direc, 'trained_model.pkl')
    trained_kge_model.load_state_dict(torch.load(path_to_model))

    entities = np.loadtxt(fname=os.path.join(data_direc, 'entities.tsv'), dtype=str, delimiter='\t')
    relations = np.loadtxt(fname=os.path.join(data_direc, 'relations.tsv'), dtype=str, delimiter='\t')

    device_name = 'cuda:0' if torch.cuda.is_available() and config[PREFERRED_DEVICE] == GPU else CPU
    device = torch.device(device_name)

    ranked_triples = make_predictions(
        kge_model=trained_kge_model,
        entities=entities,
        relations=relations,
        entity_to_id=entity_to_id,
        rel_to_id=relation_to_id,
        device=device,
    )

    np.savetxt(os.path.join(data_direc, 'predictions.tsv'), ranked_triples, fmt='%s')



# -*- coding: utf-8 -*-

"""Implementation the command line interface needed for TransD."""

from collections import OrderedDict

from pykeen.constants import (
    BATCH_SIZE, BATCH_SIZES_ERROR_MSG, BATCH_SIZES_PRINT_MSG, BATCH_SIZES_PROMPT_MSG,
    BATCH_SIZE_ERROR_MSG, BATCH_SIZE_PRINT_MSG, BATCH_SIZE_PROMPT_MSG, EMBEDDING_DIM, EMBEDDING_DIMENSIONS_ERROR_MSG,
    EMBEDDING_DIMENSIONS_PRINT_MSG, EMBEDDING_DIMENSIONS_PROMPT_MSG, ENTITIES_EMBEDDING_DIMENSION_ERROR_MSG,
    ENTITIES_EMBEDDING_DIMENSION_PRINT_MSG, ENTITIES_EMBEDDING_DIMENSION_PROMPT_MSG, EPOCHS_ERROR_MSG, EPOCHS_PRINT_MSG,
    EPOCHS_PROMPT_MSG, EPOCH_ERROR_MSG, EPOCH_PRINT_MSG, EPOCH_PROMPT_MSG, LEARNING_RATE, LEARNING_RATES_ERROR_MSG,
    LEARNING_RATES_PRINT_MSG, LEARNING_RATES_PROMPT_MSG, LEARNING_RATE_ERROR_MSG, LEARNING_RATE_PRINT_MSG,
    LEARNING_RATE_PROMPT_MSG, MARGIN_LOSS, MARGIN_LOSSES_ERROR_MSG, MARGIN_LOSSES_PRINT_MSG, MARGIN_LOSSES_PROMPT_MSG,
    MARGIN_LOSS_ERROR_MSG, MARGIN_LOSS_PRINT_MSG, MARGIN_LOSS_PROMPT_MSG, NORMS_SCORING_FUNCTION_ERROR_MSG,
    NORMS_SCORING_FUNCTION_PRINT_MSG, NORMS_SCORING_FUNCTION_PROMPT_MSG, NORM_SCORING_FUNCTION_ERROR_MSG,
    NORM_SCORING_FUNCTION_PRINT_MSG, NORM_SCORING_FUNCTION_PROMPT_MSG, NUM_EPOCHS, RELATION_EMBEDDING_DIM,
    RELATION_EMBEDDING_DIMENSIONS_ERROR_MSG, RELATION_EMBEDDING_DIMENSIONS_PRINT_MSG,
    RELATION_EMBEDDING_DIMENSIONS_PROMPT_MSG, RELATION_EMBEDDING_DIMENSION_ERROR_MSG,
    RELATION_EMBEDDING_DIMENSION_PRINT_MSG, RELATION_EMBEDDING_DIMENSION_PROMPT_MSG, SCORING_FUNCTION_NORM,
)
from pykeen.cli.utils.cli_print_msg_helper import (
    print_batch_size_message,
    print_entities_embedding_dimension_message, print_hpo_batch_sizes_message, print_hpo_embedding_dimensions_message,
    print_hpo_epochs_message, print_hpo_learning_rates_message, print_hpo_margin_losses_message,
    print_hpo_scoring_fcts_message, print_learning_rate_message, print_number_epochs_message,
    print_relations_embedding_dimension_message, print_scoring_fct_message, print_section_divider,
    print_training_margin_loss_message,
)
from pykeen.cli.utils.cli_query_helper import (
    select_float_value, select_float_values, select_integer_value,
    select_positive_integer_values,
)
from pykeen.utilities.summarize import get_config_dict


def configure_trans_d_training_pipeline(model_name):
    """Configure Trans D from pipeline.

    :param str model_name: name of the model
    :rtype: OrderedDict
    :return: configuration dictionary
    """
    config = get_config_dict(model_name)

    # Step 1: Query embedding dimension for entities
    print_entities_embedding_dimension_message()
    embedding_dimension = select_integer_value(
        print_msg=ENTITIES_EMBEDDING_DIMENSION_PRINT_MSG,
        prompt_msg=ENTITIES_EMBEDDING_DIMENSION_PROMPT_MSG,
        error_msg=ENTITIES_EMBEDDING_DIMENSION_ERROR_MSG,
    )
    config[EMBEDDING_DIM] = embedding_dimension
    print_section_divider()

    # Step 2: Query embedding dimension for relations
    print_relations_embedding_dimension_message()
    relation_embedding_dimension = select_integer_value(
        print_msg=RELATION_EMBEDDING_DIMENSION_PRINT_MSG,
        prompt_msg=RELATION_EMBEDDING_DIMENSION_PROMPT_MSG,
        error_msg=RELATION_EMBEDDING_DIMENSION_ERROR_MSG,
    )
    config[RELATION_EMBEDDING_DIM] = relation_embedding_dimension
    print_section_divider()

    # Step 2: Query margin loss
    print_training_margin_loss_message()
    magin_loss = select_float_value(
        print_msg=MARGIN_LOSS_PRINT_MSG,
        prompt_msg=MARGIN_LOSS_PROMPT_MSG,
        error_msg=MARGIN_LOSS_ERROR_MSG
    )
    config[MARGIN_LOSS] = magin_loss
    print_section_divider()

    # Step 3: Query L_p norm as scoring function
    print_scoring_fct_message()
    scoring_fct_norm = select_integer_value(print_msg=NORM_SCORING_FUNCTION_PRINT_MSG,
                                            prompt_msg=NORM_SCORING_FUNCTION_PROMPT_MSG,
                                            error_msg=NORM_SCORING_FUNCTION_ERROR_MSG)
    config[SCORING_FUNCTION_NORM] = scoring_fct_norm
    print_section_divider()

    # Step 5: Query learning rate
    print_learning_rate_message()
    learning_rate = select_float_value(
        print_msg=LEARNING_RATE_PRINT_MSG,
        prompt_msg=LEARNING_RATE_PROMPT_MSG,
        error_msg=LEARNING_RATE_ERROR_MSG
    )
    config[LEARNING_RATE] = learning_rate
    print_section_divider()

    # Step 6: Query batch size
    print_batch_size_message()
    batch_size = select_integer_value(
        print_msg=BATCH_SIZE_PRINT_MSG,
        prompt_msg=BATCH_SIZE_PROMPT_MSG,
        error_msg=BATCH_SIZE_ERROR_MSG
    )
    config[BATCH_SIZE] = batch_size
    print_section_divider()

    # Step 7: Query number of epochs
    print_number_epochs_message()
    number_epochs = select_integer_value(
        print_msg=EPOCH_PRINT_MSG,
        prompt_msg=EPOCH_PROMPT_MSG,
        error_msg=EPOCH_ERROR_MSG
    )
    config[NUM_EPOCHS] = number_epochs
    print_section_divider()

    return config


def configure_trans_d_hpo_pipeline(model_name):
    config = get_config_dict(model_name)

    # Step 1: Query embedding dimensions for entities
    print_hpo_embedding_dimensions_message()
    entity_embedding_dimensions = select_positive_integer_values(EMBEDDING_DIMENSIONS_PRINT_MSG,
                                                          EMBEDDING_DIMENSIONS_PROMPT_MSG,
                                                          EMBEDDING_DIMENSIONS_ERROR_MSG)
    config[EMBEDDING_DIM] = entity_embedding_dimensions
    print_section_divider()

    # Step 2: Query embedding dimensions for relations
    print_relations_embedding_dimension_message()
    relation_embedding_dimensions = select_positive_integer_values(
        print_msg=RELATION_EMBEDDING_DIMENSIONS_PRINT_MSG,
        prompt_msg=RELATION_EMBEDDING_DIMENSIONS_PROMPT_MSG,
        error_msg=RELATION_EMBEDDING_DIMENSIONS_ERROR_MSG
    )
    config[RELATION_EMBEDDING_DIM] = relation_embedding_dimensions
    print_section_divider()

    # Step 3: Query margin losses
    print_hpo_margin_losses_message()
    magin_loss = select_float_values(
        print_msg=MARGIN_LOSSES_PRINT_MSG,
        prompt_msg=MARGIN_LOSSES_PROMPT_MSG,
        error_msg=MARGIN_LOSSES_ERROR_MSG)
    config[MARGIN_LOSS] = magin_loss
    print_section_divider()

    # Step 4: Query L_p norms to use as scoring function
    print_hpo_scoring_fcts_message()
    scoring_fct_norm = select_positive_integer_values(
        print_msg=NORMS_SCORING_FUNCTION_PRINT_MSG,
        prompt_msg=NORMS_SCORING_FUNCTION_PROMPT_MSG,
        error_msg=NORMS_SCORING_FUNCTION_ERROR_MSG
    )
    config[SCORING_FUNCTION_NORM] = scoring_fct_norm
    print_section_divider()

    # Step 5: Query learning rate
    print_hpo_learning_rates_message()
    learning_rate = select_float_values(
        print_msg=LEARNING_RATES_PRINT_MSG,
        prompt_msg=LEARNING_RATES_PROMPT_MSG,
        error_msg=LEARNING_RATES_ERROR_MSG)
    config[LEARNING_RATE] = learning_rate
    print_section_divider()

    # Step 6: Query batch size
    print_hpo_batch_sizes_message()
    batch_size = select_positive_integer_values(
        print_msg=BATCH_SIZES_PRINT_MSG,
        prompt_msg=BATCH_SIZES_PROMPT_MSG,
        error_msg=BATCH_SIZES_ERROR_MSG)
    config[BATCH_SIZE] = batch_size
    print_section_divider()

    # Step 7: Query number of epochs
    print_hpo_epochs_message()
    number_epochs = select_positive_integer_values(
        print_msg=EPOCHS_PRINT_MSG,
        prompt_msg=EPOCHS_PROMPT_MSG,
        error_msg=EPOCHS_ERROR_MSG)
    config[NUM_EPOCHS] = number_epochs
    print_section_divider()

    return config





# -*- coding: utf-8 -*-

"""Script to compute mean rank and hits@k."""

import logging
import timeit
from dataclasses import dataclass
from typing import Callable, Dict, List, Tuple, Optional

import numpy as np
import torch

log = logging.getLogger(__name__)

DEFAULT_HITS_AT_K = [1, 3, 5, 10]


def _hash_triples(triples):
    """

    :param triples:
    :return:
    """
    return hash(tuple(triples))


def update_hits_at_k(
        hits_at_k_values: Dict[int, List[float]],
        rank_of_positive_subject_based: int,
        rank_of_positive_object_based: int) -> None:
    """Update the Hits@K dictionary for two values."""
    for k, values in hits_at_k_values.items():
        if rank_of_positive_subject_based < k:
            values.append(1.0)
        else:
            values.append(0.0)

        if rank_of_positive_object_based < k:
            values.append(1.0)
        else:
            values.append(0.0)


def _create_corrupted_triples(triple, all_entities, device):
    """

    :param triple:
    :param all_entities:
    :param device:
    :return:
    """
    candidate_entities_subject_based = np.delete(arr=all_entities, obj=triple[0:1])
    candidate_entities_subject_based = np.reshape(candidate_entities_subject_based, newshape=(-1, 1))
    candidate_entities_object_based = np.delete(arr=all_entities, obj=triple[2:3])
    candidate_entities_object_based = np.reshape(candidate_entities_object_based, newshape=(-1, 1))

    # Extract current test tuple: Either (subject,predicate) or (predicate,object)
    tuple_subject_based = np.reshape(a=triple[1:3], newshape=(1, 2))
    tuple_object_based = np.reshape(a=triple[0:2], newshape=(1, 2))

    # Copy current test tuple
    tuples_subject_based = np.repeat(a=tuple_subject_based, repeats=candidate_entities_subject_based.shape[0],
                                     axis=0)
    tuples_object_based = np.repeat(a=tuple_object_based, repeats=candidate_entities_object_based.shape[0], axis=0)

    corrupted_subject_based = np.concatenate([candidate_entities_subject_based, tuples_subject_based], axis=1)
    corrupted_subject_based = torch.tensor(corrupted_subject_based, dtype=torch.long, device=device)

    corrupted_object_based = np.concatenate([tuples_object_based, candidate_entities_object_based], axis=1)
    corrupted_object_based = torch.tensor(corrupted_object_based, dtype=torch.long, device=device)

    return corrupted_subject_based, corrupted_object_based


def _filter_corrupted_triples(corrupted_subject_based, corrupted_object_based, all_pos_triples_hashed):
    """

    :param corrupted_subject_based:
    :param corrupted_object_based:
    :param all_pos_triples_hashed:
    :return:
    """
    # TODO: Check
    corrupted_subject_based_hashed = np.apply_along_axis(_hash_triples, 1, corrupted_subject_based)
    mask = np.in1d(corrupted_subject_based_hashed, all_pos_triples_hashed, invert=True)
    mask = np.where(mask)[0]
    corrupted_subject_based = corrupted_subject_based[mask]

    corrupted_object_based_hashed = np.apply_along_axis(_hash_triples, 1, corrupted_object_based)
    mask = np.in1d(corrupted_object_based_hashed, all_pos_triples_hashed, invert=True)
    mask = np.where(mask)[0]

    if mask.size == 0:
        raise Exception("User selected filtered metric computation, but all corrupted triples exists"
                        "also a positive triples.")
    corrupted_object_based = corrupted_object_based[mask]

    return corrupted_subject_based, corrupted_object_based


def _compute_filtered_rank(
        kg_embedding_model,
        pos_triple,
        corrupted_subject_based,
        corrupted_object_based,
        device,
        all_pos_triples_hashed) -> Tuple[int, int]:
    """

    :param kg_embedding_model:
    :param pos_triple:
    :param corrupted_subject_based:
    :param corrupted_object_based:
    :param device:
    :param all_pos_triples_hashed:
    """
    corrupted_subject_based, corrupted_object_based = _filter_corrupted_triples(
        corrupted_subject_based=corrupted_subject_based,
        corrupted_object_based=corrupted_object_based,
        all_pos_triples_hashed=all_pos_triples_hashed)

    return _compute_rank(
        kg_embedding_model=kg_embedding_model,
        pos_triple=pos_triple,
        corrupted_subject_based=corrupted_subject_based,
        corrupted_object_based=corrupted_object_based,
        device=device,
        all_pos_triples_hashed=all_pos_triples_hashed,
    )


def _compute_rank(
        kg_embedding_model,
        pos_triple,
        corrupted_subject_based,
        corrupted_object_based,
        device,
        all_pos_triples_hashed=None) -> Tuple[int, int]:
    """

    :param kg_embedding_model:
    :param pos_triple:
    :param corrupted_subject_based:
    :param corrupted_object_based:
    :param device:
    :param all_pos_triples_hashed: This parameter isn't used but is necessary for compatability
    """
    scores_of_corrupted_subjects = kg_embedding_model.predict(corrupted_subject_based)
    scores_of_corrupted_objects = kg_embedding_model.predict(corrupted_object_based)

    pos_triple = np.array(pos_triple)
    pos_triple = np.expand_dims(a=pos_triple, axis=0)
    pos_triple = torch.tensor(pos_triple, dtype=torch.long, device=device)

    score_of_positive = kg_embedding_model.predict(pos_triple)

    scores_subject_based = np.append(arr=scores_of_corrupted_subjects, values=score_of_positive)
    indice_of_pos_subject_based = scores_subject_based.size - 1

    scores_object_based = np.append(arr=scores_of_corrupted_objects, values=score_of_positive)
    indice_of_pos_object_based = scores_object_based.size - 1

    _, sorted_score_indices_subject_based = torch.sort(torch.tensor(scores_subject_based, dtype=torch.float),
                                                       descending=False)
    sorted_score_indices_subject_based = sorted_score_indices_subject_based.cpu().numpy()

    _, sorted_score_indices_object_based = torch.sort(torch.tensor(scores_object_based, dtype=torch.float),
                                                      descending=False)
    sorted_score_indices_object_based = sorted_score_indices_object_based.cpu().numpy()

    # Get index of first occurrence that fulfills the condition
    rank_of_positive_subject_based = np.where(sorted_score_indices_subject_based == indice_of_pos_subject_based)[0][0]
    rank_of_positive_object_based = np.where(sorted_score_indices_object_based == indice_of_pos_object_based)[0][0]

    return (
        rank_of_positive_subject_based,
        rank_of_positive_object_based,
    )


@dataclass
class MetricResult:
    """Results from computing metrics."""
    mean_rank: float
    hits_at_k: Dict[int, float]


def compute_metric_results(all_entities,
                           kg_embedding_model,
                           mapped_train_triples,
                           mapped_test_triples,
                           device,
                           filter_neg_triples=False,
                           ks: Optional[List[int]] = None) -> MetricResult:
    """

    :param all_entities:
    :param kg_embedding_model:
    :param mapped_train_triples:
    :param mapped_test_triples:
    :param device:
    :param filter_neg_triples:
    :param ks:
    :return:
    """
    start = timeit.default_timer()

    ranks: List[int] = []
    hits_at_k_values = {
        k: []
        for k in (ks or DEFAULT_HITS_AT_K)
    }

    kg_embedding_model = kg_embedding_model.to(device)

    all_pos_triples = np.concatenate([mapped_train_triples, mapped_test_triples], axis=0)
    all_pos_triples_hashed = np.apply_along_axis(_hash_triples, 1, all_pos_triples)

    compute_rank_fct: Callable[..., Tuple[int, int]] = (
        _compute_filtered_rank
        if filter_neg_triples else
        _compute_rank
    )

    # Corrupt triples
    for pos_triple in mapped_test_triples:
        corrupted_subject_based, corrupted_object_based = _create_corrupted_triples(
            triple=pos_triple,
            all_entities=all_entities,
            device=device,
        )

        rank_of_positive_subject_based, rank_of_positive_object_based = compute_rank_fct(
            kg_embedding_model=kg_embedding_model,
            pos_triple=pos_triple,
            corrupted_subject_based=corrupted_subject_based,
            corrupted_object_based=corrupted_object_based,
            device=device,
            all_pos_triples_hashed=all_pos_triples_hashed,
        )

        ranks.append(rank_of_positive_subject_based)
        ranks.append(rank_of_positive_object_based)

        # Compute hits@k for k in {1,3,5,10}
        update_hits_at_k(
            hits_at_k_values,
            rank_of_positive_subject_based=rank_of_positive_subject_based,
            rank_of_positive_object_based=rank_of_positive_object_based,
        )

    mean_rank = float(np.mean(ranks))
    hits_at_k: Dict[int, float] = {
        k: np.mean(values)
        for k, values in hits_at_k_values.items()
    }

    stop = timeit.default_timer()
    log.debug("evaluation took %.2fs seconds", stop - start)

    return mean_rank, hits_at_k



# -*- coding: utf-8 -*-

"""Utilities for training KGE models."""

import logging
import timeit
from typing import List, Optional, Tuple

import numpy as np
import torch
import torch.optim as optim
from sklearn.utils import shuffle
from torch.nn import Module
from tqdm import trange

from pykeen.constants import *
from pykeen.kge_models import ConvE

__all__ = [
    'train_kge_model',
]

log = logging.getLogger(__name__)


def _split_list_in_batches(input_list, batch_size):
    return [input_list[i:i + batch_size] for i in range(0, len(input_list), batch_size)]


def train_kge_model(
        kge_model: Module,
        all_entities,
        learning_rate,
        num_epochs,
        batch_size,
        pos_triples,
        device,
        seed: Optional[int] = None) -> Tuple[Module, List[float]]:
    """Train the model."""
    if CONV_E_NAME == kge_model.model_name:
        return _train_conv_e_model(
            kge_model=kge_model,
            all_entities=all_entities,
            learning_rate=learning_rate,
            num_epochs=num_epochs,
            batch_size=batch_size,
            pos_triples=pos_triples,
            device=device,
            seed=seed,
        )

    # model_name in {TRANS_E_NAME, TRANS_H_NAME, TRANS_D_NAME, TRANS_R_NAME, DISTMULT_NAME, UM_NAME, SE_NAME, ERMLP_NAME, RESCAL_NAME}
    return _train_basic_model(
        kge_model=kge_model,
        all_entities=all_entities,
        learning_rate=learning_rate,
        num_epochs=num_epochs,
        batch_size=batch_size,
        pos_triples=pos_triples,
        device=device,
        seed=seed,
    )


def _train_basic_model(
        kge_model: Module,
        all_entities,
        learning_rate,
        num_epochs,
        batch_size,
        pos_triples,
        device,
        seed: Optional[int] = None) -> Tuple[Module, List[float]]:
    """"""
    if seed is not None:
        np.random.seed(seed=seed)

    kge_model = kge_model.to(device)

    optimizer = optim.SGD(kge_model.parameters(), lr=learning_rate)
    log.debug(f'****running model on {device}****')

    loss_per_epoch = []
    num_pos_triples = pos_triples.shape[0]
    num_entities = all_entities.shape[0]

    start_training = timeit.default_timer()

    for _ in trange(num_epochs, desc="Training the model (epochs)"):
        indices = np.arange(num_pos_triples)
        np.random.shuffle(indices)
        pos_triples = pos_triples[indices]
        start = timeit.default_timer()
        pos_batches = _split_list_in_batches(input_list=pos_triples, batch_size=batch_size)
        current_epoch_loss = 0.

        for i, pos_batch in enumerate(pos_batches):
            current_batch_size = len(pos_batch)
            batch_subjs = pos_batch[:, 0:1]
            batch_relations = pos_batch[:, 1:2]
            batch_objs = pos_batch[:, 2:3]

            num_subj_corrupt = len(pos_batch) // 2
            num_obj_corrupt = len(pos_batch) - num_subj_corrupt
            pos_batch = torch.tensor(pos_batch, dtype=torch.long, device=device)

            corrupted_subj_indices = np.random.choice(np.arange(0, num_entities), size=num_subj_corrupt)
            corrupted_subjects = np.reshape(all_entities[corrupted_subj_indices], newshape=(-1, 1))
            subject_based_corrupted_triples = np.concatenate(
                [corrupted_subjects, batch_relations[:num_subj_corrupt], batch_objs[:num_subj_corrupt]], axis=1)

            corrupted_obj_indices = np.random.choice(np.arange(0, num_entities), size=num_obj_corrupt)
            corrupted_objects = np.reshape(all_entities[corrupted_obj_indices], newshape=(-1, 1))

            object_based_corrupted_triples = np.concatenate(
                [batch_subjs[num_subj_corrupt:], batch_relations[num_subj_corrupt:], corrupted_objects], axis=1)

            neg_batch = np.concatenate([subject_based_corrupted_triples, object_based_corrupted_triples], axis=0)

            neg_batch = torch.tensor(neg_batch, dtype=torch.long, device=device)

            # Recall that torch *accumulates* gradients. Before passing in a
            # new instance, you need to zero out the gradients from the old instance
            optimizer.zero_grad()
            loss = kge_model(pos_batch, neg_batch)
            current_epoch_loss += (loss.item() * current_batch_size)

            loss.backward()
            optimizer.step()

        stop = timeit.default_timer()
        # log.info("Epoch %s took %s seconds \n" % (str(epoch), str(round(stop - start))))
        # Track epoch loss
        loss_per_epoch.append(current_epoch_loss / len(pos_triples))

    stop_training = timeit.default_timer()
    log.debug("training took %.2fs seconds", stop_training - start_training)

    return kge_model, loss_per_epoch


def _train_conv_e_model(
        kge_model: ConvE,
        all_entities,
        learning_rate,
        num_epochs,
        batch_size,
        pos_triples, device,
        seed: Optional[int] = None) -> Tuple[ConvE, List[float]]:
    """"""
    if seed is not None:
        np.random.seed(seed=seed)

    kge_model = kge_model.to(device)

    optimizer = optim.Adam(kge_model.parameters(), lr=learning_rate)

    loss_per_epoch: List[float] = []

    log.info('****Run Model On %s****' % str(device).upper())

    num_pos_triples = pos_triples.shape[0]
    num_entities = all_entities.shape[0]

    start_training = timeit.default_timer()

    for epoch in range(num_epochs):
        indices = np.arange(num_pos_triples)
        np.random.shuffle(indices)
        pos_triples = pos_triples[indices]
        start = timeit.default_timer()
        num_positives = batch_size // 2
        # TODO: Make sure that batch = num_pos + num_negs
        num_negatives = batch_size - num_positives

        pos_batches = _split_list_in_batches(input_list=pos_triples, batch_size=num_positives)
        current_epoch_loss = 0.

        for i in range(len(pos_batches)):
            # TODO: Remove original subject and object from entity set
            pos_batch = pos_batches[i]
            current_batch_size = len(pos_batch)
            batch_subjs = pos_batch[:, 0:1]
            batch_relations = pos_batch[:, 1:2]
            batch_objs = pos_batch[:, 2:3]

            num_subj_corrupt = len(pos_batch) // 2
            num_obj_corrupt = len(pos_batch) - num_subj_corrupt
            pos_batch = torch.tensor(pos_batch, dtype=torch.long, device=device)

            corrupted_subj_indices = np.random.choice(np.arange(0, num_entities), size=num_subj_corrupt)
            corrupted_subjects = np.reshape(all_entities[corrupted_subj_indices], newshape=(-1, 1))
            subject_based_corrupted_triples = np.concatenate(
                [corrupted_subjects, batch_relations[:num_subj_corrupt], batch_objs[:num_subj_corrupt]], axis=1)

            corrupted_obj_indices = np.random.choice(np.arange(0, num_entities), size=num_obj_corrupt)
            corrupted_objects = np.reshape(all_entities[corrupted_obj_indices], newshape=(-1, 1))

            object_based_corrupted_triples = np.concatenate(
                [batch_subjs[num_subj_corrupt:], batch_relations[num_subj_corrupt:], corrupted_objects], axis=1)

            neg_batch = np.concatenate([subject_based_corrupted_triples, object_based_corrupted_triples], axis=0)

            neg_batch = torch.tensor(neg_batch, dtype=torch.long, device=device)

            batch = np.concatenate([pos_batch, neg_batch], axis=0)
            positive_labels = np.ones(shape=(current_batch_size))
            negative_labels = np.zeros(shape=(current_batch_size))
            labels = np.concatenate([positive_labels, negative_labels], axis=0)

            batch, labels = shuffle(batch, labels, random_state=seed)

            batch = torch.tensor(batch, dtype=torch.long)
            labels = torch.tensor(labels, dtype=torch.float)

            # Recall that torch *accumulates* gradients. Before passing in a
            # new instance, you need to zero out the gradients from the old
            # instance
            optimizer.zero_grad()
            loss = kge_model(batch, labels)
            current_epoch_loss += (loss.item() * current_batch_size)

            loss.backward()
            optimizer.step()

        stop = timeit.default_timer()
        # log.info("Epoch %s took %s seconds \n" % (str(epoch), str(round(stop - start))))
        # Track epoch loss
        loss_per_epoch.append(current_epoch_loss / len(pos_triples))

    stop_training = timeit.default_timer()
    log.info("Training took %s seconds \n" % (str(round(stop_training - start_training))))

    return kge_model, loss_per_epoch



# -*- coding: utf-8 -*-

"""Utilities for getting and initializing KGE models."""

import torch
from torch import nn

from pykeen.constants import CPU, EMBEDDING_DIM, GPU, MARGIN_LOSS, NUM_ENTITIES, NUM_RELATIONS, PREFERRED_DEVICE

__all__ = [
    'BaseModule',
]


class BaseModule(nn.Module):
    """A base class for all of the models."""

    margin_ranking_loss_size_average: bool = ...

    def __init__(self, config):
        super().__init__()

        # Device selection
        self.device = torch.device('cuda:0' if torch.cuda.is_available() and config[PREFERRED_DEVICE] == GPU else CPU)

        # Loss
        self.margin_loss = config[MARGIN_LOSS]
        self.criterion = nn.MarginRankingLoss(
            margin=self.margin_loss,
            size_average=self.margin_ranking_loss_size_average,
        )

        # Entity dimensions
        self.num_entities = config[NUM_ENTITIES]
        self.num_relations = config[NUM_RELATIONS]

        self.embedding_dim = config[EMBEDDING_DIM]



# -*- coding: utf-8 -*-

"""Implementation the command line interface needed for TransE."""

from typing import Dict

from pykeen.constants import (
    BATCH_SIZE, BATCH_SIZES_ERROR_MSG, BATCH_SIZES_PRINT_MSG, BATCH_SIZES_PROMPT_MSG, BATCH_SIZE_ERROR_MSG,
    BATCH_SIZE_PRINT_MSG, BATCH_SIZE_PROMPT_MSG, EMBEDDING_DIM, EMBEDDING_DIMENSIONS_ERROR_MSG,
    EMBEDDING_DIMENSIONS_PRINT_MSG, EMBEDDING_DIMENSIONS_PROMPT_MSG, EMBEDDING_DIMENSION_ERROR_MSG,
    EMBEDDING_DIMENSION_PRINT_MSG, EMBEDDING_DIMENSION_PROMPT_MSG, EPOCHS_ERROR_MSG, EPOCHS_PRINT_MSG,
    EPOCHS_PROMPT_MSG, EPOCH_ERROR_MSG, EPOCH_PRINT_MSG, EPOCH_PROMPT_MSG, LEARNING_RATE, LEARNING_RATES_ERROR_MSG,
    LEARNING_RATES_PRINT_MSG, LEARNING_RATES_PROMPT_MSG, LEARNING_RATE_ERROR_MSG, LEARNING_RATE_PRINT_MSG,
    LEARNING_RATE_PROMPT_MSG, MARGIN_LOSS, MARGIN_LOSSES_ERROR_MSG, MARGIN_LOSSES_PRINT_MSG, MARGIN_LOSSES_PROMPT_MSG,
    MARGIN_LOSS_ERROR_MSG, MARGIN_LOSS_PRINT_MSG, MARGIN_LOSS_PROMPT_MSG, NORMS_SCORING_FUNCTION_ERROR_MSG,
    NORMS_SCORING_FUNCTION_PRINT_MSG, NORMS_SCORING_FUNCTION_PROMPT_MSG, NORM_SCORING_FUNCTION_ERROR_MSG,
    NORM_SCORING_FUNCTION_PRINT_MSG, NORM_SCORING_FUNCTION_PROMPT_MSG, NUM_EPOCHS, SCORING_FUNCTION_NORM,
    WEIGHTS_SOFT_CONSTRAINT_TRANS_H_ERROR_MSG, WEIGHTS_SOFT_CONSTRAINT_TRANS_H_PRINT_MSG,
    WEIGHTS_SOFT_CONSTRAINT_TRANS_H_PROMPT_MSG, WEIGHT_SOFT_CONSTRAINT_TRANS_H,
)
from pykeen.cli.utils.cli_print_msg_helper import (
    print_batch_size_message, print_embedding_dimension_info_message, print_hpo_batch_sizes_message,
    print_hpo_embedding_dimensions_message, print_hpo_epochs_message, print_hpo_learning_rates_message,
    print_hpo_margin_losses_message, print_hpo_scoring_fcts_message, print_hpo_trans_h_soft_constraints_weights_message,
    print_learning_rate_message, print_number_epochs_message, print_scoring_fct_message, print_section_divider,
    print_training_embedding_dimension_message, print_training_margin_loss_message,
    print_trans_h_soft_constraints_weight_message,
)
from pykeen.cli.utils.cli_query_helper import (
    select_float_value, select_float_values, select_integer_value, select_positive_integer_values,
)
from pykeen.utilities.summarize import get_config_dict


def configure_trans_h_training_pipeline(model_name: str) -> Dict:
    """Prompt the user to configure Trans H from pipeline."""
    config = get_config_dict(model_name)

    # Step 1: Query embedding dimension
    print_training_embedding_dimension_message()
    print_embedding_dimension_info_message()
    embedding_dimension = select_integer_value(
        print_msg=EMBEDDING_DIMENSION_PRINT_MSG,
        prompt_msg=EMBEDDING_DIMENSION_PROMPT_MSG,
        error_msg=EMBEDDING_DIMENSION_ERROR_MSG,
    )
    config[EMBEDDING_DIM] = embedding_dimension
    print_section_divider()

    # Step 2: Query margin loss
    print_training_margin_loss_message()
    magin_loss = select_float_value(
        print_msg=MARGIN_LOSS_PRINT_MSG,
        prompt_msg=MARGIN_LOSS_PROMPT_MSG,
        error_msg=MARGIN_LOSS_ERROR_MSG,
    )
    config[MARGIN_LOSS] = magin_loss
    print_section_divider()

    # Step 3: Query L_p norm as scoring function
    print_scoring_fct_message()
    scoring_fct_norm = select_integer_value(
        print_msg=NORM_SCORING_FUNCTION_PRINT_MSG,
        prompt_msg=NORM_SCORING_FUNCTION_PROMPT_MSG,
        error_msg=NORM_SCORING_FUNCTION_ERROR_MSG
    )
    config[SCORING_FUNCTION_NORM] = scoring_fct_norm
    print_section_divider()

    # Step 4: Query weight for the soft constraints
    print_trans_h_soft_constraints_weight_message()
    soft_constraints_weight = select_float_value(
        print_msg=WEIGHTS_SOFT_CONSTRAINT_TRANS_H_PRINT_MSG,
        prompt_msg=WEIGHTS_SOFT_CONSTRAINT_TRANS_H_PROMPT_MSG,
        error_msg=WEIGHTS_SOFT_CONSTRAINT_TRANS_H_ERROR_MSG,
    )
    config[WEIGHT_SOFT_CONSTRAINT_TRANS_H] = soft_constraints_weight
    print_section_divider()

    # Step 5: Query learning rate
    print_learning_rate_message()
    learning_rate = select_float_value(
        print_msg=LEARNING_RATE_PRINT_MSG,
        prompt_msg=LEARNING_RATE_PROMPT_MSG,
        error_msg=LEARNING_RATE_ERROR_MSG,
    )
    config[LEARNING_RATE] = learning_rate
    print_section_divider()

    # Step 6: Query batch size
    print_batch_size_message()
    batch_size = select_integer_value(
        print_msg=BATCH_SIZE_PRINT_MSG,
        prompt_msg=BATCH_SIZE_PROMPT_MSG,
        error_msg=BATCH_SIZE_ERROR_MSG,
    )
    config[BATCH_SIZE] = batch_size
    print_section_divider()

    # Step 7: Query number of epochs
    print_number_epochs_message()
    number_epochs = select_integer_value(
        print_msg=EPOCH_PRINT_MSG,
        prompt_msg=EPOCH_PROMPT_MSG,
        error_msg=EPOCH_ERROR_MSG,
    )
    config[NUM_EPOCHS] = number_epochs
    print_section_divider()

    return config


def configure_trans_h_hpo_pipeline(model_name):
    config = get_config_dict(model_name)

    # Step 1: Query embedding dimensions
    print_hpo_embedding_dimensions_message()
    embedding_dimensions = select_positive_integer_values(
        EMBEDDING_DIMENSIONS_PRINT_MSG,
        EMBEDDING_DIMENSIONS_PROMPT_MSG,
        EMBEDDING_DIMENSIONS_ERROR_MSG,
    )
    config[EMBEDDING_DIM] = embedding_dimensions
    print_section_divider()

    # Step 2: Query margin loss
    print_hpo_margin_losses_message()
    magin_loss = select_float_values(
        print_msg=MARGIN_LOSSES_PRINT_MSG,
        prompt_msg=MARGIN_LOSSES_PROMPT_MSG,
        error_msg=MARGIN_LOSSES_ERROR_MSG,
    )

    config[MARGIN_LOSS] = magin_loss
    print_section_divider()

    # Step 3: Query L_p norms to use as scoring function
    print_hpo_scoring_fcts_message()
    scoring_fct_norm = select_positive_integer_values(
        print_msg=NORMS_SCORING_FUNCTION_PRINT_MSG,
        prompt_msg=NORMS_SCORING_FUNCTION_PROMPT_MSG,
        error_msg=NORMS_SCORING_FUNCTION_ERROR_MSG,
    )
    config[SCORING_FUNCTION_NORM] = scoring_fct_norm
    print_section_divider()

    # Step 4: Query weight for the soft constraints
    print_hpo_trans_h_soft_constraints_weights_message()
    soft_constraints_weight = select_float_values(
        print_msg=WEIGHTS_SOFT_CONSTRAINT_TRANS_H_PRINT_MSG,
        prompt_msg=WEIGHTS_SOFT_CONSTRAINT_TRANS_H_PROMPT_MSG,
        error_msg=WEIGHTS_SOFT_CONSTRAINT_TRANS_H_ERROR_MSG,
    )
    config[WEIGHT_SOFT_CONSTRAINT_TRANS_H] = soft_constraints_weight
    print_section_divider()

    # Step 5: Query learning rate
    print_hpo_learning_rates_message()
    learning_rate = select_float_values(
        print_msg=LEARNING_RATES_PRINT_MSG,
        prompt_msg=LEARNING_RATES_PROMPT_MSG,
        error_msg=LEARNING_RATES_ERROR_MSG,
    )
    config[LEARNING_RATE] = learning_rate
    print_section_divider()

    # Step 6: Query batch size
    print_hpo_batch_sizes_message()
    batch_sizes = select_positive_integer_values(
        print_msg=BATCH_SIZES_PRINT_MSG,
        prompt_msg=BATCH_SIZES_PROMPT_MSG,
        error_msg=BATCH_SIZES_ERROR_MSG,
    )
    config[BATCH_SIZE] = batch_sizes
    print_section_divider()

    # Step 7: Query number of epochs
    print_hpo_epochs_message()
    number_epochs = select_positive_integer_values(
        print_msg=EPOCHS_PRINT_MSG,
        prompt_msg=EPOCHS_PROMPT_MSG,
        error_msg=EPOCHS_ERROR_MSG,
    )
    config[NUM_EPOCHS] = number_epochs
    print_section_divider()

    return config



# -*- coding: utf-8 -*-

"""Implementation of structured model (SE)."""

import logging

import numpy as np
import torch
import torch.autograd
from torch import nn

from pykeen.constants import NORM_FOR_NORMALIZATION_OF_ENTITIES, SCORING_FUNCTION_NORM, SE_NAME
from pykeen.kge_models.base import BaseModule

__all__ = ['StructuredEmbedding']

log = logging.getLogger(__name__)


class StructuredEmbedding(BaseModule):
    """An implementation of Structured Embedding (SE) [bordes2011]_.

    This model projects different matrices for each relation head and tail entity.

    .. [bordes2011] Bordes, A., *et al.* (2011). `Learning Structured Embeddings of Knowledge Bases
                    <http://www.aaai.org/ocs/index.php/AAAI/AAAI11/paper/download/3659/3898>`_. AAAI. Vol. 6. No. 1.
    """

    model_name = SE_NAME
    margin_ranking_loss_size_average: bool = True

    def __init__(self, config):
        super().__init__(config)

        # Embeddings
        self.l_p_norm_entities = config[NORM_FOR_NORMALIZATION_OF_ENTITIES]
        self.scoring_fct_norm = config[SCORING_FUNCTION_NORM]
        self.entity_embeddings = nn.Embedding(self.num_entities, self.embedding_dim)
        self.m_left_rel_embeddings = nn.Embedding(self.num_relations, self.embedding_dim * self.embedding_dim)
        self.m_right_rel_embeddings = nn.Embedding(self.num_relations, self.embedding_dim * self.embedding_dim)

        self._initialize()

    def _initialize(self):
        """

        :return:
        """
        lower_bound = -6 / np.sqrt(self.embedding_dim)
        upper_bound = 6 / np.sqrt(self.embedding_dim)
        nn.init.uniform_(self.entity_embeddings.weight.data, a=lower_bound, b=upper_bound)
        nn.init.uniform_(self.m_left_rel_embeddings.weight.data, a=lower_bound, b=upper_bound)

        norms = torch.norm(self.m_left_rel_embeddings.weight, p=2, dim=1).data
        self.m_left_rel_embeddings.weight.data = self.m_left_rel_embeddings.weight.data.div(
            norms.view(self.num_relations, 1).expand_as(self.m_left_rel_embeddings.weight))

    def _compute_loss(self, pos_scores, neg_scores):
        """

        :param pos_scores:
        :param neg_scores:
        :return:
        """

        y = np.repeat([-1], repeats=pos_scores.shape[0])
        y = torch.tensor(y, dtype=torch.float, device=self.device)

        # Scores for the psotive and negative triples
        pos_scores = torch.tensor(pos_scores, dtype=torch.float, device=self.device)
        neg_scores = torch.tensor(neg_scores, dtype=torch.float, device=self.device)

        loss = self.criterion(pos_scores, neg_scores, y)

        return loss

    def _compute_scores(self, projected_head_embs, projected_tail_embs):
        """

        :param h_embs:
        :param r_embs:
        :param t_embs:
        :return:
        """

        # Subtract the vector element wise
        difference = projected_head_embs - projected_tail_embs

        distances = torch.norm(difference, dim=1, p=self.scoring_fct_norm).view(size=(-1,))

        return distances

    def _project_entities(self, entity_embs, projection_matrix_embs):
        entity_embs = entity_embs.unsqueeze(-1)

        projected_entity_embs = torch.matmul(projection_matrix_embs, entity_embs)

        return projected_entity_embs

    def predict(self, triples):
        """

        :param head:
        :param relation:
        :param tail:
        :return:
        """
        # triples = torch.tensor(triples, dtype=torch.long, device=self.device)

        heads = triples[:, 0:1]
        relations = triples[:, 1:2]
        tails = triples[:, 2:3]

        head_embs = self.entity_embeddings(heads).view(-1, self.embedding_dim)
        tail_embs = self.entity_embeddings(tails).view(-1, self.embedding_dim)

        m_left_r_embs = self.m_left_rel_embeddings(relations).view(-1, self.embedding_dim, self.embedding_dim)
        m_right_r_embs = self.m_right_rel_embeddings(relations).view(-1, self.embedding_dim, self.embedding_dim)

        projected_head_embs = self._project_entities(entity_embs=head_embs,
                                                     projection_matrix_embs=m_left_r_embs)
        projected_tails_embs = self._project_entities(entity_embs=tail_embs,
                                                      projection_matrix_embs=m_right_r_embs)

        scores = self._compute_scores(projected_head_embs=projected_head_embs,
                                      projected_tail_embs=projected_tails_embs)

        return scores.detach().cpu().numpy()

    def forward(self, batch_positives, batch_negatives):
        """

        :param batch_positives:
        :param batch_negatives:
        :return:
        """

        # Normalise embeddings of entities
        norms = torch.norm(self.entity_embeddings.weight, p=self.l_p_norm_entities, dim=1).data
        self.entity_embeddings.weight.data = self.entity_embeddings.weight.data.div(
            norms.view(self.num_entities, 1).expand_as(self.entity_embeddings.weight))

        pos_heads = batch_positives[:, 0:1]
        pos_relations = batch_positives[:, 1:2]
        pos_tails = batch_positives[:, 2:3]

        neg_heads = batch_negatives[:, 0:1]
        neg_relations = batch_negatives[:, 1:2]
        neg_tails = batch_negatives[:, 2:3]

        pos_h_embs = self.entity_embeddings(pos_heads).view(-1, self.embedding_dim)
        pos_m_left_r_embs = self.m_left_rel_embeddings(pos_relations).view(-1, self.embedding_dim, self.embedding_dim)
        pos_m_right_r_embs = self.m_right_rel_embeddings(pos_relations).view(-1, self.embedding_dim, self.embedding_dim)

        pos_t_embs = self.entity_embeddings(pos_tails).view(-1, self.embedding_dim)

        neg_h_embs = self.entity_embeddings(neg_heads).view(-1, self.embedding_dim)
        neg_m_left_r_embs = self.m_left_rel_embeddings(neg_relations).view(-1, self.embedding_dim, self.embedding_dim)
        neg_m_right_r_embs = self.m_right_rel_embeddings(neg_relations).view(-1, self.embedding_dim, self.embedding_dim)
        neg_t_embs = self.entity_embeddings(neg_tails).view(-1, self.embedding_dim)

        projected_pos_head_embs = self._project_entities(entity_embs=pos_h_embs,
                                                         projection_matrix_embs=pos_m_left_r_embs)
        projected_pos_tails_embs = self._project_entities(entity_embs=pos_t_embs,
                                                          projection_matrix_embs=pos_m_right_r_embs)

        projected_neg_head_embs = self._project_entities(entity_embs=neg_h_embs,
                                                         projection_matrix_embs=neg_m_left_r_embs)
        projected_neg_tails_embs = self._project_entities(entity_embs=neg_t_embs,
                                                          projection_matrix_embs=neg_m_right_r_embs)

        pos_scores = self._compute_scores(projected_head_embs=projected_pos_head_embs,
                                          projected_tail_embs=projected_pos_tails_embs)

        neg_scores = self._compute_scores(projected_head_embs=projected_neg_head_embs,
                                          projected_tail_embs=projected_neg_tails_embs)

        loss = self._compute_loss(pos_scores=pos_scores, neg_scores=neg_scores)

        return loss



# -*- coding: utf-8 -*-

""""""

__all__ = [
    'transform_freebase_to_interal_format',
    'transform_wn_18_to_interal_format',
]


def transform_freebase_to_interal_format(data_path_in, data_path_out):
    _transform_subj_obj_pred_to_internal_foramt(data_path_in, data_path_out)


def transform_wn_18_to_interal_format(data_path_in, data_path_out):
    _transform_subj_obj_pred_to_internal_foramt(data_path_in, data_path_out)


def _transform_subj_obj_pred_to_internal_foramt(data_path_in, data_path_out):
    with open(data_path_in, 'r', encoding='utf-8') as f1, open(data_path_out, 'w', encoding='utf-8') as f2:
        lines = f1.readlines()
        f2.write('@Comment@ Subject Predicate Object\n')
        for line in lines:
            parts = line.strip().split('\t')
            subject = parts[0]
            object = parts[1]
            predicate = parts[2]
            assert len(parts) == 3
            parts = [subject, predicate, object]
            f2.write('\t'.join(parts) + '\n')



# -*- coding: utf-8 -*-

import logging
from typing import Dict, Optional, Tuple

import numpy as np

__all__ = [
    'create_mapped_triples',
    'create_mappings',
]

log = logging.getLogger(__name__)


def create_mapped_triples(triples: np.ndarray,
                          entity_to_id: Optional[Dict[int, str]] = None,
                          rel_to_id: Optional[Dict[int, str]] = None) -> np.ndarray:
    """"""
    if entity_to_id is None or rel_to_id is None:
        entity_to_id, rel_to_id = create_mappings(triples)

    subject_column = np.vectorize(entity_to_id.get)(triples[:, 0:1])
    relation_column = np.vectorize(rel_to_id.get)(triples[:, 1:2])
    object_column = np.vectorize(entity_to_id.get)(triples[:, 2:3])
    triples_of_ids = np.concatenate([subject_column, relation_column, object_column], axis=1)

    triples_of_ids = np.array(triples_of_ids, dtype=np.long)
    # Note: Unique changes the order
    return np.unique(ar=triples_of_ids, axis=0), entity_to_id, rel_to_id


def create_mappings(triples: np.ndarray) -> Tuple[Dict[str, int], Dict[str, int]]:
    """"""
    entities = np.unique(np.ndarray.flatten(np.concatenate([triples[:, 0:1], triples[:, 2:3]])))
    relations = np.unique(np.ndarray.flatten(triples[:, 1:2]).tolist())

    entity_to_id: Dict[int, str] = {
        value: key
        for key, value in enumerate(entities)
    }

    rel_to_id: Dict[int, str] = {
        value: key
        for key, value in enumerate(relations)
    }

    return entity_to_id, rel_to_id



# -*- coding: utf-8 -*-

"""Utilities for the command line interface."""

from .conv_e_cli import (
    configure_conv_e_hpo_pipeline, configure_conv_e_training_pipeline,  # noqa: 401; noqa: 401
)
from .distmult_cli import (
    configure_distmult_hpo_pipeline, configure_distmult_training_pipeline,  # noqa: 401; noqa: 401
)
from .ermlp_cli import (
    configure_ermlp_hpo_pipeline, configure_ermlp_training_pipeline,  # noqa: 401; noqa: 401
)
from .rescal_cli import (
    configure_rescal_hpo_pipeline, configure_rescal_training_pipeline,  # noqa: 401; noqa: 401
)
from .structured_embedding_cli import (
    configure_se_hpo_pipeline, configure_se_training_pipeline,  # noqa: 401; noqa: 401
)
from .trans_d_cli import (
    configure_trans_d_hpo_pipeline, configure_trans_d_training_pipeline,  # noqa: 401; noqa: 401
)
from .trans_e_cli import (
    configure_trans_e_hpo_pipeline, configure_trans_e_training_pipeline,  # noqa: 401; noqa: 401
)
from .trans_h_cli import (
    configure_trans_h_hpo_pipeline, configure_trans_h_training_pipeline,  # noqa: 401; noqa: 401
)
from .trans_r_cli import (
    configure_trans_r_hpo_pipeline,
    configure_trans_r_training_pipeline,  # noqa: 401; noqa: 401
)
from .unstructured_model_cli import (
    configure_um_hpo_pipeline, configure_um_training_pipeline,  # noqa: 401; noqa: 401
)




# -*- coding: utf-8 -*-

"""PyKEEN's command line interface."""

import json
import os

import click

from pykeen.cli.prompt import prompt_config
from pykeen.predict import start_predictions_pipeline
from pykeen.run import run
from pykeen.utilities.summarize import summarize_results


@click.command()
@click.version_option()
@click.option('-c', '--config', type=click.File(), help='A PyKEEN JSON configuration file')
def main(config):
    """PyKEEN: A software for training and evaluating knowledge graph embeddings."""
    if config is not None:
        config = json.load(config)
    else:
        config = prompt_config()

    run(config)


@click.command()
@click.option('-m', '--model-directory', type=click.Path(file_okay=False, dir_okay=True))
@click.option('-d', '--data-directory', type=click.Path(file_okay=False, dir_okay=True))
def predict(model_directory: str, data_directory: str):
    """Predict new links based on trained model."""
    start_predictions_pipeline(model_directory, data_directory)


@click.command()
@click.option('-d', '--directory', type=click.Path(file_okay=False, dir_okay=True), default=os.getcwd())
@click.option('-o', '--output', type=click.File('w'))
def summarize(directory: str, output):
    """Summarize contents of training and evaluation."""
    summarize_results(directory, output)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Script for initializing the knowledge graph embedding models."""

from typing import Dict, Mapping

import torch.optim as optim

from pykeen.constants import (
    ADAGRAD_OPTIMIZER_NAME, ADAM_OPTIMIZER_NAME, LEARNING_RATE, OPTMIZER_NAME, SGD_OPTIMIZER_NAME,
)

__all__ = [
    'OPTIMIZERS',
    'get_optimizer',
]

OPTIMIZERS: Mapping = {
    SGD_OPTIMIZER_NAME: optim.SGD,
    ADAGRAD_OPTIMIZER_NAME: optim.Adagrad,
    ADAM_OPTIMIZER_NAME: optim.Adam,
}


def get_optimizer(config: Dict, kge_model):
    """Get an optimizer for the given knowledge graph embedding model."""
    optimizer_name = config[OPTMIZER_NAME]
    optimizer_cls = OPTIMIZERS.get(optimizer_name)

    if optimizer_cls is None:
        raise ValueError(f'invalid optimizer name: {optimizer_name}')

    parameters = filter(lambda p: p.requires_grad, kge_model.parameters())

    return optimizer_cls(parameters, lr=config[LEARNING_RATE])



# -*- coding: utf-8 -*-

"""Utilities for predicting links."""

from itertools import product

import numpy as np
import torch

from pykeen.utilities.triples_creation_utils import create_mapped_triples

__all__ = [
    'create_triples',
    'make_predictions',
]


def create_triples(entity_pairs, relation):
    subjects = entity_pairs[:, 0:1]
    objects = entity_pairs[:, 1:2]
    relation_repeated = np.reshape(np.repeat(relation, repeats=subjects.shape[0]), newshape=(-1, 1))

    triples = np.concatenate([subjects, relation_repeated, objects], axis=1)

    return triples


def make_predictions(kge_model, entities, relations, entity_to_id, rel_to_id, device):
    """

    :param kge_model: Trained KG model
    :param candidates: numpy array with two columns: 1.) Entites 2.) Relations
    :return:
    """
    all_entity_pairs = np.array(list(product(entities, entities)))

    if relations.size == 1:
        all_triples = create_triples(entity_pairs=all_entity_pairs, relation=relations)
    else:
        all_triples = create_triples(entity_pairs=all_entity_pairs, relation=relations[0])

        for relation in relations[1:]:
            triples = create_triples(entity_pairs=all_entity_pairs, relation=relation)
            all_triples = np.append(all_triples, triples, axis=0)

    mapped_triples, _, _ = create_mapped_triples(all_triples, entity_to_id=entity_to_id, rel_to_id=rel_to_id)

    mapped_triples = torch.tensor(mapped_triples, dtype=torch.long, device=device)

    id_to_entity = {value: key for key, value in entity_to_id.items()}
    id_to_relation = {value: key for key, value in rel_to_id.items()}
    subject_column = np.vectorize(id_to_entity.get)(mapped_triples[:, 0:1])
    predicate_column = np.vectorize(id_to_relation.get)(mapped_triples[:, 1:2])
    object_column = np.vectorize(id_to_entity.get)(mapped_triples[:, 2:3])

    triples_ordered_by_ids = np.concatenate([subject_column, predicate_column, object_column], axis=1)

    predicted_scores = kge_model.predict(mapped_triples)

    _, sorted_indices = torch.sort(torch.tensor(predicted_scores, dtype=torch.float),
                                   descending=False)

    sorted_indices = sorted_indices.cpu().numpy()

    ranked_triples = triples_ordered_by_ids[sorted_indices, :]

    subs = np.reshape(ranked_triples[:, 0:1], newshape=(-1))
    preds = np.reshape(ranked_triples[:, 1:2], newshape=(-1))
    objs = np.reshape(ranked_triples[:, 2:3], newshape=(-1))

    indices = np.where(subs != objs)[0]

    ranked_scores = np.reshape(predicted_scores[sorted_indices], newshape=(-1, 1))
    ranked_triples = np.concatenate([ranked_triples, ranked_scores], axis=1)

    return ranked_triples[indices, :]



# -*- coding: utf-8 -*-

"""Test that all models can be instantiated."""

import unittest

from pykeen.constants import *
from pykeen.kge_models import TransE


class TestModelInstantiation(unittest.TestCase):
    """Test that all models can be instantiated."""

    def test_instantiate_trans_e(self):
        """Test that TransE can be instantiated."""
        trans_e = TransE(config={
            NUM_ENTITIES: 5,
            NUM_RELATIONS: 5,
            EMBEDDING_DIM: 5,
            NORM_FOR_NORMALIZATION_OF_ENTITIES: 2,
            SCORING_FUNCTION_NORM: 1,
            MARGIN_LOSS: 4,
        })
        # TODO @mehdi type checking for parameters
        # TODO @mehdi @cthoyt refactor reused code for TransXXX models to base class
        self.assertIsNotNone(trans_e)
